import os
import plistlib
import warnings

from numbers_parser.document import Document

DEFAULT_NUMBERS_INSTALL_PATH = "/Applications/Numbers.app"
VERSION_PLIST_PATH = "Contents/version.plist"
SUPPORTED_NUMBERS_VERSIONS = [
    "10.3",
    "11.0",
]

# Don't print the source line
formatwarning_old = warnings.formatwarning
warnings.formatwarning = lambda message, category, filename, lineno, line=None: \
    formatwarning_old(message, category, filename, lineno, line='')

def check_installed_numbers_version():
    try:
        fp = open(os.path.join(DEFAULT_NUMBERS_INSTALL_PATH, VERSION_PLIST_PATH), "rb")
    except IOError:
        return None
    version_dict = plistlib.load(fp)
    installed_version = version_dict["CFBundleShortVersionString"]
    if installed_version not in SUPPORTED_NUMBERS_VERSIONS:
        warnings.warn(
            f"Numbers version {installed_version} not tested with this version"
        )

check_installed_numbers_version()



import array
import importlib
import os
import re
import sys
import struct
import zipfile

from enum import Enum
from datetime import datetime, timedelta

from numbers_parser.containers import ItemsList, ObjectStore, NumbersError
from numbers_parser.generated import TSTArchives_pb2 as TSTArchives


class UnsupportedError(NumbersError):
    """Raised for unsupported file format features"""

    pass


class Document:
    def __init__(self, filename):
        self._object_store = ObjectStore(filename)

    def sheets(self):
        refs = [o.identifier for o in self._object_store[1].sheets]
        self._sheets = Sheets(self._object_store, refs)
        return self._sheets


class Sheets(ItemsList):
    def __init__(self, object_store, refs):
        super(Sheets, self).__init__(object_store, refs, Sheet)


class Sheet:
    def __init__(self, object_store, sheet_id):
        self._sheet_id = sheet_id
        self._object_store = object_store
        self._sheet = object_store[sheet_id]
        self.name = self._sheet.name

    def tables(self):
        table_ids = self._object_store.find_refs("TableInfoArchive")
        table_refs = [
            self._object_store[table_id].tableModel.identifier
            for table_id in table_ids
            if self._object_store[table_id].super.parent.identifier == self._sheet_id
        ]
        self._tables = Tables(self._object_store, table_refs)
        return self._tables


import binascii


class Tables(ItemsList):
    def __init__(self, object_store, refs):
        super(Tables, self).__init__(object_store, refs, Table)


class Table:
    def __init__(self, object_store, table_id):
        super(Table, self).__init__()
        self._object_store = object_store
        self._table = object_store[table_id]
        self._table_id = table_id
        self.name = self._table.table_name

        bds = self._table.base_data_store
        self._row_headers = [
            h.numberOfCells for h in object_store[bds.rowHeaders.buckets[0].identifier].headers
        ]
        self._column_headers = [
            h.numberOfCells for h in object_store[bds.columnHeaders.identifier].headers
        ]
        self._tile_ids = [t.tile.identifier for t in bds.tiles.tiles]

    @property
    def data(self):
        row_infos = []
        for tile_id in self._tile_ids:
            row_infos += self._object_store[tile_id].rowInfos

        storage_version = self._object_store[tile_id].storage_version
        if storage_version != 5:
            raise UnsupportedError(f"Unsupported row info version {storage_version}")

        storage_buffers = [
            extract_cell_data(
                r.cell_storage_buffer, r.cell_offsets, self.num_cols, r.has_wide_offsets
            )
            for r in row_infos
        ]
        storage_buffers_pre_bnc = [
            extract_cell_data(
                r.cell_storage_buffer_pre_bnc,
                r.cell_offsets_pre_bnc,
                self.num_cols,
                r.has_wide_offsets,
            )
            for r in row_infos
        ]

        data = []
        num_rows = sum([self._object_store[t].numrows for t in self._tile_ids])
        for row_num in range(num_rows):
            row = []
            for col_num in range(self.num_cols):
                if col_num < len(storage_buffers[row_num]):
                    storage_buffer = storage_buffers[row_num][col_num]
                else:
                    # Rest of row is empty cells
                    row.append(None)
                    continue

                if col_num < len(storage_buffers_pre_bnc[row_num]):
                    storage_buffer_pre_bnc = storage_buffers_pre_bnc[row_num][col_num]
                else:
                    storage_buffer_pre_bnc = None

                if storage_buffer is None:
                    row.append(None)
                else:
                    cell_type = storage_buffer[1]
                    if cell_type == TSTArchives.emptyCellValueType:
                        cell_value = None
                    elif cell_type == TSTArchives.numberCellType:
                        if storage_buffer_pre_bnc is None:
                            cell_value = None
                        else:
                            cell_value = struct.unpack("<d", storage_buffer_pre_bnc[-12:-4])[0]
                    elif cell_type == TSTArchives.textCellType:
                        key = struct.unpack("<i", storage_buffer[12:16])[0]
                        cell_value = self._table_string(key)
                    elif cell_type == TSTArchives.dateCellType:
                        if storage_buffer_pre_bnc is None:
                            cell_value = None
                        else:
                            seconds = struct.unpack("<d", storage_buffer_pre_bnc[-12:-4])[0]
                            cell_value = datetime(2001, 1, 1) + timedelta(seconds=seconds)
                    elif cell_type == TSTArchives.boolCellType:
                        d = struct.unpack("<d", storage_buffer[12:20])[0]
                        cell_value = d > 0.0
                    elif cell_type == TSTArchives.durationCellType:
                        cell_value = struct.unpack("<d", storage_buffer[12:20])[0]
                    elif cell_type == TSTArchives.formulaErrorCellType:
                        cell_value = "*ERROR*"
                    elif cell_type == 9:
                        print(
                            f"[{row_num},{col_num}]: cell type {cell_type}, buffer:",
                            binascii.hexlify(storage_buffer, sep=":"),
                            "bnc_buffer:",
                            binascii.hexlify(storage_buffer_pre_bnc, sep=":"),
                        )
                        cell_value = "*FORMULA*"
                    elif cell_type == 10:
                        if storage_buffer_pre_bnc is None:
                            cell_value = None
                        else:
                            cell_value = struct.unpack("<d", storage_buffer_pre_bnc[-12:-4])[0]
                    else:
                        print(
                            f"[{row_num},{col_num}]: unknown cell type {cell_type}, buffer:",
                            binascii.hexlify(storage_buffer, sep=":"),
                            "bnc_buffer:",
                            binascii.hexlify(storage_buffer_pre_bnc, sep=":"),
                        )
                        raise UnsupportedError(
                            f"Unsupport cell type {cell_type} @{self.name}:({row_num},{col_num})"
                        )

                    row.append(cell_value)
            data.append(row)

        return data

    @property
    def num_rows(self):
        return len(self._row_headers)

    @property
    def num_cols(self):
        return len(self._column_headers)

    def _table_string(self, key):
        if not hasattr(self, "_table_strings"):
            strings_id = self._table.base_data_store.stringTable.identifier
            self._table_strings = {x.key: x.string for x in self._object_store[strings_id].entries}
        return self._table_strings[key]


def extract_cell_data(storage_buffer, offsets, num_cols, has_wide_offsets):
    offsets = array.array("h", offsets).tolist()
    if has_wide_offsets:
        offsets = [o * 4 for o in offsets]

    data = []
    for col_num in range(num_cols):
        if col_num >= len(offsets):
            break

        start = offsets[col_num]
        if start < 0:
            data.append(None)
            continue

        if col_num == (len(offsets) - 1):
            end = len(storage_buffer)
        else:
            # Get next offset past current one that is not -1
            #  https://stackoverflow.com/questions/19502378/
            idx = next((i for i, x in enumerate(offsets[col_num + 1 :]) if x >= 0), None)
            if idx is None:
                end = len(storage_buffer)
            else:
                end = offsets[col_num + idx + 1]
        data.append(storage_buffer[start:end])

    return data



from builtins import zip
from builtins import str
from builtins import object

import sys
import struct
import snappy
import traceback
from functools import partial

from numbers_parser.mapping import ID_NAME_MAP, NAME_CLASS_MAP

from google.protobuf.internal.encoder import _VarintBytes
from google.protobuf.internal.decoder import _DecodeVarint32
from google.protobuf.json_format import MessageToDict, ParseDict
from google.protobuf.message import EncodeError

from numbers_parser.generated.TSPArchiveMessages_pb2 import ArchiveInfo


MAX_FLOAT = 340282346638528859811704183484516925440.000000000000000000


class IWAFile(object):
    def __init__(self, chunks, filename=None):
        self.chunks = chunks
        self.filename = filename

    @classmethod
    def from_buffer(cls, data, filename=None):
        try:
            chunks = []
            while data:
                chunk, data = IWACompressedChunk.from_buffer(data, filename)
                chunks.append(chunk)

            return cls(chunks, filename)
        except Exception as e:
            if filename:
                raise ValueError("Failed to deserialize " + filename) from e
            else:
                raise

    @classmethod
    def from_dict(cls, _dict):
        return cls([IWACompressedChunk.from_dict(chunk) for chunk in _dict["chunks"]])

    def to_dict(self):
        try:
            return {"chunks": [chunk.to_dict() for chunk in self.chunks]}
        except Exception as e:
            if self.filename:
                raise ValueError("Failed to serialize " + self.filename) from e
            else:
                raise

    def to_buffer(self):
        return b"".join([chunk.to_buffer() for chunk in self.chunks])


class IWACompressedChunk(object):
    def __init__(self, archives):
        self.archives = archives

    def __eq__(self, other):
        return self.archives == other.archives

    @classmethod
    def _decompress_all(cls, data):
        while data:
            header = data[:4]

            first_byte = header[0]
            if not isinstance(first_byte, int):
                first_byte = ord(first_byte)

            if first_byte != 0x00:
                raise ValueError(
                    "IWA chunk does not start with 0x00! (found %x)" % first_byte
                )

            unpacked = struct.unpack_from("<I", bytes(header[1:]) + b"\x00")
            length = unpacked[0]
            chunk = data[4 : 4 + length]
            data = data[4 + length :]

            try:
                yield snappy.uncompress(chunk)
            except Exception:
                # Try to see if this data isn't compressed in the first place.
                # If this data is still compressed, parsing it as Protobuf
                # will almost definitely fail anyways.
                yield chunk

    @classmethod
    def from_buffer(cls, data, filename=None):
        data = b"".join(cls._decompress_all(data))
        archives = []
        while data:
            archive, data = IWAArchiveSegment.from_buffer(data, filename)
            archives.append(archive)
        return cls(archives), None

    @classmethod
    def from_dict(cls, _dict):
        return cls([IWAArchiveSegment.from_dict(d) for d in _dict["archives"]])

    def to_dict(self):
        return {"archives": [archive.to_dict() for archive in self.archives]}

    def to_buffer(self):
        uncompressed = b"".join([archive.to_buffer() for archive in self.archives])
        payloads = []
        while uncompressed:
            payloads.append(snappy.compress(uncompressed[:65536]))
            uncompressed = uncompressed[65536:]
        return b"".join(
            [
                b"\x00" + struct.pack("<I", len(payload))[:3] + payload
                for payload in payloads
            ]
        )


class ProtobufPatch(object):
    def __init__(self, data):
        self.data = data

    def __eq__(self, other):
        return self.data == other.data

#    def __repr__(self):
#        return "<%s %s>" % (self.__class__.__name__, self.data)

    def to_dict(self):
        return message_to_dict(self.data)

    @classmethod
    def FromString(cls, message_info, proto_klass, data):
        if len(message_info.diff_field_path.path) != 1:
            # print(
            #     "warning: "
            #     + proto_klass.__name__
            #     + ": can't process diff_field_path = "
            #     + str(len(message_info.diff_field_path.path)),
            #     file=sys.stderr,
            # )
            return cls(proto_klass.FromString(data))

        if message_info.fields_to_remove:
            print(
                str(proto_klass) + ": can't process fields_to_remove", file=sys.stderr
            )
            return cls(proto_klass.FromString(data))

        for diff_path in message_info.diff_field_path.path:
            if diff_path not in proto_klass.DESCRIPTOR.fields_by_number:
                print(str(proto_klass) + ": " + str(diff_path) + " path not found", file=sys.stderr)
                return cls(proto_klass.FromString(data))
            else:
                patched_field = proto_klass.DESCRIPTOR.fields_by_number[diff_path]
                print(str(proto_klass) + ": patching", str(patched_field), file=sys.stderr)
                field_message_class = NAME_CLASS_MAP[patched_field.message_type.full_name]
                return cls(field_message_class.FromString(data))

    def SerializeToString(self):
        return self.data.SerializePartialToString()


class IWAArchiveSegment(object):
    def __init__(self, header, objects):
        self.header = header
        self.objects = objects

    def __eq__(self, other):
        return self.header == other.header and self.objects == other.objects

#    def __repr__(self):
#        return "<%s identifier=%s objects=%s>" % (
#            self.__class__.__name__,
#            self.header.identifier,
#            repr(self.objects).replace("\n", " ").replace("  ", " "),
#        )

    @classmethod
    def from_buffer(cls, buf, filename=None):
        archive_info, payload = get_archive_info_and_remainder(buf)
        if not repr(archive_info):
            raise ValueError("Segment doesn't seem to start with an ArchiveInfo!")

        payloads = []

        n = 0
        for message_info in archive_info.message_infos:
            try:
                if message_info.type == 0 and archive_info.should_merge and payloads:
                    base_message = archive_info.message_infos[
                        message_info.base_message_index
                    ]
                    klass = partial(
                        ProtobufPatch.FromString,
                        message_info,
                        ID_NAME_MAP[base_message.type],
                    )
                else:
                    klass = ID_NAME_MAP[message_info.type]
            except KeyError:
                raise NotImplementedError(
                    "Don't know how to parse Protobuf message type "
                    + str(message_info.type)
                )
            try:
                message_payload = payload[n : n + message_info.length]
                if hasattr(klass, "FromString"):
                    output = klass.FromString(message_payload)
                else:
                    output = klass(message_payload)
            except Exception as e:
                raise ValueError(
                    "Failed to deserialize %s payload of length %d: %s"
                    % (klass, message_info.length, e)
                )
            payloads.append(output)
            n += message_info.length

        return cls(archive_info, payloads), payload[n:]

    @classmethod
    def from_dict(cls, _dict):
        header = dict_to_header(_dict["header"])
        objects = []
        for message_info, o in zip(header.message_infos, _dict["objects"]):
            if message_info.diff_field_path and "_pbtype" not in o:
                base_message_info = header.message_infos[
                    message_info.base_message_index
                ]
                message_class = ID_NAME_MAP[base_message_info.type]
                objects.append(ProtobufPatch(message_class, o))
            else:
                objects.append(dict_to_message(o))
        return cls(header, objects)

    def to_dict(self):
        return {
            "header": header_to_dict(self.header),
            "objects": [message_to_dict(message) for message in self.objects],
        }

    def to_buffer(self):
        # Each message_info as part of the header needs to be updated
        # so that its length matches the object contained within.
        for obj, message_info in zip(self.objects, self.header.message_infos):
            try:
                object_length = len(obj.SerializeToString())
                provided_length = message_info.length
                if object_length != provided_length:
                    message_info.length = object_length
            except EncodeError as e:
                raise ValueError(
                    "Failed to encode object: %s\nObject: '%s'\nMessage info: %s"
                    % (e, repr(obj), message_info)
                )
        return b"".join(
            [_VarintBytes(self.header.ByteSize()), self.header.SerializeToString()]
            + [obj.SerializeToString() for obj in self.objects]
        )


def message_to_dict(message):
    if hasattr(message, "to_dict"):
        return message.to_dict()
    output = MessageToDict(message)
    output["_pbtype"] = type(message).DESCRIPTOR.full_name
    return output


def header_to_dict(message):
    output = message_to_dict(message)
    for message_info in output["messageInfos"]:
        del message_info["length"]
    return output


def _work_around_protobuf_max_float_handling(_dict):
    """
    protobuf==3.13.0 includes a check to make sure that any `float` value
    parsed from a Python dictionary must be less than FLOAT_MAX. Unfortunately,
    when parsing from YAML, serialized FLOAT_MAX values round to slightly over FLOAT_MAX.
    This is a nasty hack to fix that.
    """
    for k, v in _dict.items():
        if isinstance(v, dict):
            _work_around_protobuf_max_float_handling(v)
        elif isinstance(v, float):
            if v > MAX_FLOAT:
                _dict[k] = MAX_FLOAT
    return _dict


def dict_to_message(_dict):
    _type = _dict["_pbtype"]
    del _dict["_pbtype"]
    _dict = _work_around_protobuf_max_float_handling(_dict)
    return ParseDict(_dict, NAME_CLASS_MAP[_type](), ignore_unknown_fields=True)


def dict_to_header(_dict):
    for message_info in _dict["messageInfos"]:
        # set a dummy length value that we'll overwrite later
        message_info["length"] = 0
    return dict_to_message(_dict)


def get_archive_info_and_remainder(buf):
    msg_len, new_pos = _DecodeVarint32(buf, 0)
    n = new_pos
    msg_buf = buf[n : n + msg_len]
    n += msg_len
    return ArchiveInfo.FromString(msg_buf), buf[n:]



from zipfile import ZipFile, BadZipFile
from numbers_parser.codec import IWAFile
from pathlib import Path

import os


class NumbersError(Exception):
    """Base class for other exceptions"""

    pass


class FileError(NumbersError):
    """Raised for IO and other OS errors"""

    pass


class FileFormatError(NumbersError):
    """Raised for parsing errors during file load"""

    pass


class ItemsList:
    def __init__(self, objects, refs, item_class):
        self._item_name = item_class.__name__.lower()
        self._items = [item_class(objects, _) for _ in refs]

    def __getitem__(self, key: int):
        if type(key) == int:
            if key < 0 or key >= len(self._items):
                raise IndexError(f"index {key} out of range")
            return self._items[key]
        elif type(key) == str:
            for item in self._items:
                if item.name == key:
                    return item
            raise KeyError(f"no {self._item_name} named '{key}'")
        else:
            t = type(key).__name__
            raise LookupError(f"invalid index type {t}")

    def __len__(self):
        return len(self._items)


class ObjectStore:
    def __init__(self, filename):
        objects = {}
        if os.path.isdir(filename):
            if os.path.isfile(os.path.join(filename, "Index.zip")):
                try:
                    zipf = ZipFile(os.path.join(filename, "Index.zip"))
                except BadZipFile as e:
                    raise FileError(f"{index_zip}: " + str(e))

                iwa_files = filter(lambda x: x.endswith(".iwa"), zipf.namelist())
                for iwa_filename in iwa_files:
                    # TODO: LZFSE compressed according to /usr/bin/file
                    if "OperationStorage" in iwa_filename:
                        continue
                    contents = zipf.read(iwa_filename)
                    objects.update(extract_iwa_archives(contents, iwa_filename))
            else:
                iwa_files = list(Path(filename).rglob("*.iwa"))
                for iwa_filename in iwa_files:
                    f = open(iwa_filename, "rb")
                    contents = f.read()
                    objects.update(extract_iwa_archives(contents, iwa_filename))
        else:
            try:
                zipf = ZipFile(filename)
            except BadZipFile as e:
                raise FileError(f"{filename}: " + str(e))

            iwa_files = filter(lambda x: x.endswith(".iwa"), zipf.namelist())
            for iwa_filename in iwa_files:
                contents = zipf.read(iwa_filename)
                objects.update(extract_iwa_archives(contents, iwa_filename))

        self._object_store = objects

    def __getitem__(self, key):
        return self._object_store[key]

    def __len__(self):
        return len(self._object_store)

    def find_refs(self, ref_name):
        refs = [
            k for k, v in self._object_store.items() if type(v).__name__ == ref_name
        ]
        return refs

    def find_objects(self, ref_name, class_name):
        refs = self.find_refs(ref_name)
        class_ = getattr(importlib.import_module(__name__), class_name)
        return [class_(self, obj_id) for obj_id in refs]


def extract_iwa_archives(contents, iwa_filename):
    objects = {}
    try:
        iwaf = IWAFile.from_buffer(contents, iwa_filename)
    except Exception as e:
        raise FileFormatError(f"{iwa_filename}: invalid IWA file {iwa_filename}") from e

    if len(iwaf.chunks) != 1:
        raise FileFormatError(f"{iwa_filename}: chunk count != 1 in {iwa_filename}")
    for archive in iwaf.chunks[0].archives:
        if len(archive.objects) == 0:
            raise FileFormatError(f"{iwa_filename}: no objects in {iwa_filename}")

        identifier = archive.header.identifier
        if identifier in objects:
            raise FileFormatError(f"{iwa_filename}: duplicate reference {identifier}")

        if len(archive.objects) == 1:
            objects[identifier] = archive.objects[0]
        else:
            #  print(f"warning: {iwa_filename}: found", len(archive.objects), "objects")
            pass

    return objects



from setuptools import setup, find_packages

setup(
    name="numbers-parser",
    version="1.0.2",
    author="Jon Connell",
    author_email="python@figsandfudge.com",
    description="Package to read data from Apple Numbers spreadsheets",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/masaccio/numbers-parser",
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=[
        "protobuf",
        "pytest",
        "python-snappy",
        "PyYAML",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
